/*

    This file is part of gerbvQt.
    (c) Kurganov Alexander, 2016 me@sx107.ru

    gerbvQt is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Foobar is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gerbvQt.  If not, see <http://www.gnu.org/licenses/>.

*/


#include "gerbv.h"
#include "gerbvQt.h"
#include <iostream>
#include <cairo.h>
#include <QFlags>
#include <QTemporaryFile>
#include <QSettings>
#include <QtCore/qmath.h>

static double  scaleFactor = 25.4 / 0.051; // One pixel is 0.051mm
static double  border   = 50.0;                       // 50mm border
static bool    positive = false, small_drils = false; // to se da nastavit
static bool    mirrory  = true;
static QString suffix (".png");
static double info_max_x, info_min_x, info_max_y, info_min_y; 
/////////////////////////////////////////////////////////////////////
using namespace std;

/**
 * Do adresáře $HOME/.config/Kizarm dáme konfigurační soubor QtGerb.conf s obsahem něco jako
[General]
negative=true
small_drils=true
mirrory=false
suffix=.bmp
scaleFactor=498.04
border=10 
 * */
static QSettings settings ("Kizarm", "QtGerb");
static bool Settings () {
  positive    =!settings.value("negative")   .toBool();
  small_drils = settings.value("small_drils").toBool();
  mirrory     = settings.value("mirrory")    .toBool();
  QString s   = settings.value("suffix")   .toString();
  const QStringList sl = {".bmp",".jpg",".png",".ppm",".xbm",".xpm"};
  for (const QString & e : sl) {
    if (e == s) { suffix = s; break; }
  }
  bool ok = false;
  double d = 0.0;
  d = settings.value("scaleFactor").toDouble(&ok);
  if (ok) scaleFactor = d;
  d = settings.value("border"     ).toDouble(&ok);
  if (ok) border = d;
  
  info_max_x = settings.value("info_max_x").toDouble(&ok);
  if (!ok) return false;
  info_min_x = settings.value("info_min_x").toDouble(&ok);
  if (!ok) return false;
  info_max_y = settings.value("info_max_y").toDouble(&ok);
  if (!ok) return false;
  info_min_y = settings.value("info_min_y").toDouble(&ok);
  if (!ok) return false;
  return true;
}
static void SettingsSave () {
  settings.setValue ("info_max_x", info_max_x);
  settings.setValue ("info_min_x", info_min_x);
  settings.setValue ("info_max_y", info_max_y);
  settings.setValue ("info_min_y", info_min_y);
}
/*
 * Nejjednodušší způsob jak zmenšit díry je použít regexp na jejich průměry - všechny
 */
static bool addDrillFile (gerbv_project_t * gProject, const char * name, const bool mark = true) {
  QTemporaryFile drlf;    // Docasny soubor se vytvari vzdy - je to tak jednodussi
  QFile inpf (name);
  drlf.open ();
  if (!inpf.open (QIODevice::ReadOnly  | QIODevice::Text)) return false;
  QString diam;
  diam.sprintf ("\\1%g", 0.4); // prumer 0.4 mm (minimum = via)
  while (!inpf.atEnd()) {
    QString line = inpf.readLine();
    if (mark)   line.replace (QRegExp("(T[0-9]+C)[0-9]+\\.[0-9]+"), diam);
    drlf.write (line.toLatin1());
  }
  inpf.close();
  drlf.close();
  gerbv_open_layer_from_filename (gProject, drlf.fileName().toLatin1().data());
  return true;
}

int main (int argc, char** argv) {
  if (argc < 2) {cerr << "Usage: ./gerb anygerberfile [drillfile]" << endl; return 1;}
  
  bool dims_ok = Settings();
  bool export_drill = false;
  //Create a project and load one file in it
  gerbv_project_t * mainProject = gerbv_create_project();
  gerbv_open_layer_from_filename (mainProject, argv [1]);
  if (argc > 2) {
    if (!strcmp (argv[2], "-w")) dims_ok = false;
    else export_drill = addDrillFile  (mainProject, argv [2], small_drils);
  }
  gerbv_image_info_t* gInfo = mainProject->file[0]->image->info;

  // gInfo->polarity = GERBV_POLARITY_NEGATIVE; // tady se to neda pouzit - buhvi jak to autor myslel
  
  if (!dims_ok) { // pokud je druhý parametr -w, zapiš rozměry
    info_max_x = gInfo->max_x, info_min_x = gInfo->min_x;
    info_max_y = gInfo->max_y, info_min_y = gInfo->min_y;
    SettingsSave();
  }
  //Compute the image size
  int size_x = qCeil ( (info_max_x - info_min_x) * scaleFactor) + border * 2;
  int size_y = qCeil ( (info_max_y - info_min_y) * scaleFactor) + border * 2;

  //Compute the render info
  gerbv_render_info_t RenderInfo;

  RenderInfo.renderType = GERBV_RENDER_TYPE_CAIRO_HIGH_QUALITY;

  //Same as the image size.
  RenderInfo.displayWidth  = size_x;
  RenderInfo.displayHeight = size_y;

  //ScaleFactor
  RenderInfo.scaleFactorX = scaleFactor;
  RenderInfo.scaleFactorY = scaleFactor;

  //Shift
  RenderInfo.lowerLeftX = info_min_x - border / scaleFactor;
  RenderInfo.lowerLeftY = info_min_y - border / scaleFactor;

  //Draw using Qt
//const QSize display (1024,768);       // buď plný rozměr obrazovky v pixelech
  const QSize display (size_x, size_y); // nebo jen rozměr vrstvy - pak musí být offset (0,0)
  QImage qtimage (display, QImage::Format_ARGB32);
  qtimage.fill(Qt::black);
  QImage l1 (size_x, size_y, QImage::Format_ARGB32), l2 (l1);

  //Fun starts here
  gerbvQt gqt;

  const QColor white_np (255,255,255, 255), black_np (0,0,0, 255);
  const QColor fgb = positive ? black_np : white_np;
  const QColor fgw = positive ? white_np : black_np;
  //Set fg/bg colors
  gqt.setForegroundColor (fgb);                     // cerna/bila neprusvitna
  gqt.setBackgroundColor (QColor (255,255,255, 0)); // bila  - uplne prusvitna

  //Some options (see gerbv.h)
  gqt.setDrawingMode (gerbvQt::dm_CompositionModes);
  gqt.setFillFullDevice (true);
  gqt.setInitFill       (true);
  gqt.setRenderHints (QPainter::RenderHints (0));

  //Draw it!
  gqt.renderLayerToQt   (&l1, mainProject->file[0], &RenderInfo);
  if (export_drill) {
    gqt.setForegroundColor (fgw);  // neprusvitna - tecky der
    gqt.renderLayerToQt (&l2, mainProject->file[1], &RenderInfo);
  }

  // spojit vrstvy - kompozici nechapu, ale lze to takhle obejit
  // vrstva do samostatneho QImage a pak je vykreslit.
  QPainter painter (&qtimage);
  const QColor bg = positive ? white_np : black_np;
  if (mirrory) {
    const QMatrix m (-1, 0 , 0, 1, display.width(),  0);  // pripadne zrcadleni podle x
    painter.setMatrix (m);
  }
  QPoint offset (0, 0);   // v pixelech vůči levému, v případě zrcadlení pravému hornímu rohu
#if 0
  // Takto by mohla fungovat panelizace
  while (offset.y() + size_y < display.height()) {
    offset.setX(10);
    while (offset.x() + size_x < display.width()) {
      QRect rect (offset, QSize (size_x, size_y));
      painter.fillRect (rect, bg);
      painter.drawImage   (offset, l1);
      if (export_drill)
        painter.drawImage (offset, l2);
      offset.setX(offset.x() + size_x + 1);
    }
    offset.setY(offset.y() + size_y + 1);
  }
#else
  QRect rect (offset, QSize (size_x, size_y));
  painter.fillRect (rect, bg);
  painter.drawImage   (offset, l1);
  if (export_drill)
    painter.drawImage (offset, l2);
#endif
  //Save the image
  QString fs = QString::fromUtf8 (argv [1]);
  fs.truncate  (fs.lastIndexOf ('.'));
  // 1 bitova kopie - neni jasne jaky format bude potreba, asi se bude muset upravit
  QImage cp = qtimage.convertToFormat (QImage::Format_Mono, Qt::AutoColor);
  cp.save (fs + suffix);               // podle pripony typ - .bmp je bitmapa
  
  /* bez toho neni nutne pouzit libcairo
  //Draw the image using cairo
  cairo_surface_t* surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, size_x, size_y);
  cairo_t* cr = cairo_create (surface);
  cairo_set_source_rgba (cr, 1,1,1,1);
  cairo_paint (cr);

  GdkColor blackColor = {0, 0, 0, 0};
  mainProject->file[0]->color = blackColor;

  cairo_push_group (cr);
  gerbv_render_layer_to_cairo_target (cr, mainProject->file[0], &RenderInfo);
  cairo_pop_group_to_source (cr);
  cairo_paint_with_alpha (cr, 1);
  cairo_surface_flush (surface);

  cairo_surface_write_to_png (surface, "cairo.png");
  */
  return 0;
}
